from typing import Iterable, Sequence, TypeVar, Callable

A = TypeVar('A')

def find_idx(p: Callable[[A], bool], xs: Iterable[A]) -> int | None:
  for i, x in enumerate(xs):
    if p(x):
      return i

def find(p: Callable[[A], bool], xs: Iterable[A]) -> A | None:
  for x in xs:
    if p(x):
      return x
    
def find_last_idx(p: Callable[[A], bool], xs: Sequence[A]) -> int | None:
  for i, x in enumerate(reversed(xs)):
    if p(x):
      return len(xs) - i - 1

def find_last(p: Callable[[A], bool], xs: Sequence[A]) -> A | None:
  for x in reversed(xs):
    if p(x):
      return x


from .either import Either, Left, Right, IsLeft
from .narrowing import is_left, is_right
from .funcs import safe, sequence, filter, filter_lefts, maybe, take_while, unsafe, get_or
from .pydantic import validate, validate_json
from .do_notation import do

__all__ = [
  'Either', 'Left', 'Right', 'IsLeft', 'is_left', 'is_right',
  'safe', 'sequence', 'filter', 'filter_lefts', 'maybe', 'take_while', 'unsafe',
  'validate', 'validate_json', 'do', 'get_or',
]


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from dataclasses import dataclass
from typing_extensions import TypeVar, Callable, Generic
from haskellian import Monad

A = TypeVar('A')
B = TypeVar('B')

@dataclass
class Thunk(Monad[A], Generic[A]):

  supplier: Callable[[], A]
  value: A | None = None

  def fmap(self, f: Callable[[A], B]) -> 'Thunk[B]':
    return Thunk(lambda: f(self()))
  
  def bind(self, f: Callable[[A], 'Thunk[B]']) -> 'Thunk[B]':
    def _supplier():
      value = self()
      thunk = f(value)
      value2 = thunk()
      return value2
    return Thunk(_supplier)
  
  def ap(self, f: 'Thunk[Callable[[A], B]]') -> 'Thunk[B]':
    return super().ap(f) # type: ignore

  def __or__(self, f: Callable[[A], B]) -> 'Thunk[B]':
    return self.fmap(f)
  
  def __call__(self) -> A:
    if self.value is None:
      self.value = self.supplier()
    return self.value
  
  @classmethod
  def of(cls, value: A) -> 'Thunk[A]':
    return cls(lambda: value)


from inspect import Signature, Parameter

def add_kw(sig: Signature, name: str, type: type) -> Signature:
  """Adds a keyword-only parameter to a signature."""
  params = list(sig.parameters.values())
  params.append(Parameter(name, Parameter.KEYWORD_ONLY, default=None, annotation=type))
  return sig.replace(parameters=params)


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Callable, Generic, TypeVar
from abc import ABC, abstractmethod
from .functor import Functor

A = TypeVar('A', covariant=True)
B = TypeVar('B')

class Applicative(Functor[A], ABC, Generic[A]):
  
  @classmethod
  @abstractmethod
  def pure(cls, value: B) -> 'Applicative[B]':
    ...

  @abstractmethod
  def ap(self, f: 'Applicative[Callable[[A], B]]') -> 'Applicative[B]':
    ...

  def fmap(self, f: Callable[[A], B]) -> 'Applicative[B]':
    return self.ap(self.pure(f))



from haskellian import DEBUG_IMPORTS, iter as I
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, TypeVarTuple, Iterable, Callable, Unpack
import itertools

A = TypeVar('A')
B = TypeVar('B')
As = TypeVarTuple('As')

def fst(t: tuple[A, Unpack[As]]) -> A:
	"""`fst((a, _)) = a`"""
	x, *_ = t
	return x

def snd(t: tuple[A, B, Unpack[As]]) -> B:
	"""`snd((_, b)) = b`"""
	_, x, *_ = t
	return x

def head(xs: Iterable[A]) -> A | None:
  """`head([x, *_]) = x`"""
  for x in xs:
    return x

@I.lift
def tail(xs: Iterable[A]) -> Iterable[A]:
	"""`tail([_, *xs]) = xs`"""
	return skip(1, xs)

def last(xs: Iterable[A]) -> A | None:
  """`head([x, *_]) = x`"""
  ys = list(xs)
  if len(ys) > 0:
    return ys[-1]

@I.lift
def take(n: int | None, xs: Iterable[A]) -> Iterable[A]:
	"""`take(n, [x1, ..., xn, *_]) = [x1, ..., xn]`"""
	return itertools.islice(xs, n)

@I.lift
def take_while(pred: Callable[[A], bool], xs: Iterable[A]) -> Iterable[A]:
	for x in xs:
		if not pred(x):
			return
		yield x

@I.lift
def drop_while(pred: Callable[[A], bool], xs: Iterable[A]) -> Iterable[A]:
	started = False
	for x in xs:
		if not started and not pred(x):
			started = True
		if started:
			yield x

@I.lift  
def skip(n: int, xs: Iterable[A]) -> Iterable[A]:
	"""`skip(n, [x1, ..., xn, *xs]) = xs`"""
	return itertools.islice(xs, n, None)

@I.lift
def pad(n: int, fill: B, xs: Iterable[A]) -> Iterable[A|B]:
	"""Pads `xs` to length `n`, appending `fill`s as needed"""
	i = 0
	for i, x in enumerate(xs):
		yield x
		
	for _ in range(i+1, n):
		yield fill


from typing_extensions import TypeVar, Callable, Iterable, overload
from .either import Either, Left, Right

A = TypeVar('A')
R = TypeVar('R')
L = TypeVar('L')
Err = TypeVar('Err', bound=Exception)

def safe(func: Callable[[], R], Exc: type[Err] = Exception) -> 'Either[Err, R]':
    try:
      return Right(func())
    except Exc as e:
      return Left(e)
    
def maybe(x: R | None) -> 'Either[None, R]':
  return Left(None) if x is None else Right(x)

@overload
def get_or(default: R) -> Callable[[Either[L, R]], R]: ...
@overload
def get_or(default: A) -> Callable[[Either[L, R]], R | A]: ...
def get_or(default: A) -> Callable[[Either[L, R]], R | A]:
  return lambda e: e.get_or(default)

def unsafe(x: Either[L, R]) -> R:
  return x.unsafe()
    
def sequence(eithers: Iterable[Either[L, R]]) -> Either[list[L], list[R]]:
  """List of lefts if any (thus with length in `[1, len(eithers)]`), otherwise list of all rights (with length `len(eithers)`)"""
  lefts: list[L] = []
  rights: list[R] = []
  for x in eithers:
    x.match(lefts.append, rights.append)
  return Right(rights) if lefts == [] else Left(lefts)

def filter(eithers: Iterable[Either[L, R]]) -> Iterable[R]:
  for e in eithers:
    match e:
      case Right(value):
        yield value

def filter_lefts(eithers: Iterable[Either[L, R]]) -> Iterable[L]:
  for e in eithers:
    match e:
      case Left(err):
        yield err

def take_while(eithers: Iterable[Either[L, R]]) -> Iterable[R]:
  for e in eithers:
    match e:
      case Right(x):
        yield x
      case _:
        return


from typing_extensions import Callable, Generic, TypeVar, Mapping, Iterable, Any
from haskellian import Monad, dicts as D, iter as I

K = TypeVar('K')
K2 = TypeVar('K2')
V = TypeVar('V')
V2 = TypeVar('V2')

class Dict(Monad[tuple[K, V]], dict[K, V], Generic[K, V]):
  """A monadic, fluent `dict`"""
  
  def __repr__(self):
    return f'Dict({super().__repr__()})'

  @classmethod
  def of(cls, x: tuple[K, V]):
    return cls(dict([x]))
  
  def flatmap(self, f: Callable[[K, V], Mapping[K2, V2]]) -> 'Dict[K2, V2]':
    return D.flatmap(f, self)
  
  bind = flatmap # type: ignore

  def flatmap_k(self, f: Callable[[K], Mapping[K2, V]]) -> 'Dict[K2, V]':
    return D.flatmap_k(f, self)
  
  def flatmap_v(self, f: Callable[[V], Mapping[K, V2]]) -> 'Dict[K, V2]':
    return D.flatmap_v(f, self)
  
  def map(self, f: Callable[[K, V], V2]) -> 'Dict[K, V2]':
    return D.map(f, self)
  
  def map_v(self, f: Callable[[V], V2]) -> 'Dict[K, V2]':
    return D.map_v(f, self)
  
  def map_k(self, f: Callable[[K], K2]) -> 'Dict[K2, V]':
    return D.map_k(f, self)
  
  def filter(self, p):
    return D.filter(p, self)
  
  def filter_k(self, f):
    return D.filter_k(f, self)
  
  def filter_v(self, f):
    return D.filter_v(f, self)
  
  def zip(self, f: Callable[[V], Iterable[V2]] = lambda x: x) -> 'I.Iter[Dict[K, V2]]': # type: ignore
    """By default, `zip({ 'a': [1, 2, 3], 'b': [4, 5, 6] }) == [{'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 3, 'b': 6}]`
    - Note: perhaps `f` is useful for something else, but it's just a dummy to make the types work.
    """
    return D.zip(self.map(f)) # type: ignore

  def unpack(self, *keys: K) -> tuple[V, ...]:
    return D.unpack(self, *keys)
  
  def evolve(self, mappers: Mapping[K, Callable[[V], V2]]) -> 'Dict[K, V|V2]':
    return D.evolve(mappers, self)
  
  def group_by(self, f: Callable[[K, V], K2]) -> 'Dict[K2, Dict[K, V]]':
    return D.group_by(lambda t: f(*t), self.items()).map_v(Dict)
  
  def sorted(self, key: Callable[[K, V], Any]) -> list[V]:
    return [v for _, v in sorted(self.items(), key=lambda t: key(*t))]


from typing_extensions import Callable, TypeVar, ParamSpec, Generic, overload, Coroutine, Awaitable
from inspect import iscoroutinefunction
from functools import wraps
from .either import Either, Left, Right, IsLeft

L = TypeVar('L')
R = TypeVar('R')
P = ParamSpec('P')

class do(Generic[L]):
  """Lift a function with do notation
  ```
  @do[LeftType]()
  def myfn(e: Either[LeftType, int]):
    value = e.unsafe()
    return value + 1

  myfn(Left('err')) # Left('err')
  myfn(Right(1)) # Right(2)
  ```
  """
  @overload
  def __call__(self, fn: Callable[P, Coroutine[R, None, None]]) -> Callable[P, Awaitable[Either[L, R]]]:
    ...
  @overload
  def __call__(self, fn: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[Either[L, R]]]:
    ...
  @overload
  def __call__(self, fn: Callable[P, R]) -> Callable[P, Either[L, R]]:
    ...
  
  def __call__(self, fn): # type: ignore
    if iscoroutinefunction(fn):
      @wraps(fn)
      async def _wrapper(*args: P.args, **kwargs: P.kwargs):
        try:
          return Right(await fn(*args, **kwargs))
        except IsLeft as e:
          return Left(e.value)
      return _wrapper
    else:
      @wraps(fn)
      def wrapper(*args: P.args, **kwargs: P.kwargs):
        try:
          return Right(fn(*args, **kwargs))
        except IsLeft as e:
          return Left(e.value)
      return wrapper


from haskellian import DEBUG_IMPORTS, iter as I, promise as P, Promise
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, Callable, Awaitable, Mapping, Iterable, overload, ParamSpec, Coroutine, Any
from functools import wraps
from inspect import isawaitable
import asyncio

A = TypeVar('A')
B = TypeVar('B')
A = TypeVar('A')
Ps = ParamSpec('Ps')

def run(coro: Callable[Ps, Coroutine[Any, Any, A]]) -> Callable[Ps, A]:
  @wraps(coro)
  def _wrapped(*args: Ps.args, **kwargs: Ps.kwargs):
    return asyncio.run(coro(*args, **kwargs))
  return _wrapped

async def then(f: Callable[[A], B], x: Awaitable[A]) -> B:
  return f(await x)

async def bind(f: Callable[[A], Awaitable[B]], x: Awaitable[A]) -> B:
  return await f(await x)

@P.lift
async def of(x: A) -> A:
  return x

@P.lift
async def delay(secs: float):
  await asyncio.sleep(secs)

@P.lift
async def wait(x: A | Awaitable[A]) -> A:
  return await x if isawaitable(x) else x # type: ignore

@overload
def all(xs: Iterable[Awaitable[A]]) -> Promise[list[A]]: ...
@overload
def all(xs: Mapping[str, Awaitable[A]]) -> Promise[dict[str, A]]: ...

@P.lift
async def all(xs):
  if isinstance(xs, Mapping):
    keys, values = I.unzip(xs.items())
    results = await asyncio.gather(*values)
    return dict(zip(keys, results))
  else:
    return await asyncio.gather(*xs)

@P.lift
async def all2d(xxs: Iterable[Iterable[Awaitable[A]]]) -> list[list[A]]:
  return await all([all(xs) for xs in xxs])


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Iterable, TypeVar, Callable
from itertools import chain
import haskellian.iter as I

A = TypeVar('A')
B = TypeVar('B')

def isiterable(x, str_iterable = False, bytes_iterable = False) -> bool:
  """Is `x` iterable?
  - `str_iterable`, `bytes_iterable`: whether `str` and `bytes` are considered iterable
  """
  return isinstance(x, Iterable) and \
    (not isinstance(x, str) or str_iterable) and \
    (not isinstance(x, bytes) or bytes_iterable)

@I.lift
def flatten(xs: Iterable[Iterable[A]]) -> Iterable[A]:
  """Single-level list flattening"""
  return chain.from_iterable(xs)

@I.lift
def flatmap(f: Callable[[A], Iterable[B]], xs: Iterable[A]) -> Iterable[B]:
  return flatten(map(f, xs))

@I.lift
def range(start: int = 0, end: int | None = None, step: int = 1) -> Iterable[int]:
  """Like `range`, but possibly infinite (if `end is None`)"""
  i = start
  while end is None or i < end:
    yield i
    i += step

@I.lift
def tap(f: Callable[[A], None], xs: Iterable[A]) -> Iterable[A]:
	for x in xs:
		f(x)
		yield x


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, TypeGuard
from .either import Either, Left, Right

L = TypeVar('L')
R = TypeVar('R')

def is_left(either: Either[L, R]) -> TypeGuard[Left[L, R]]:
  return either.tag == 'left'

def is_right(either: Either[L, R]) -> TypeGuard[Right[L, R]]:
  return either.tag == 'right'


from .lifting import lift
from .basics import isiterable, flatmap, flatten, range, tap
from .zipping import unzip, uncons, pairwise
from .slicing import fst, snd, head, tail, last, skip, take, take_while, drop_while, pad
from .indexing import at, pick
from .batching import batch, split, shard, lazy_batch, lazy_shard
from .nested import ndmap, ndrange, ndenumerate, ndflat
from .transposing import transpose, transpose_ragged
from .curried import map, filter, max, min, sorted
from .searching import find_idx, find_last_idx, find, find_last
from .maps import pluck
from .iter import Iter

__all__ = [
  'isiterable', 'flatmap', 'flatten', 'range',
  'pluck',
  'unzip', 'uncons', 'pairwise', 'tap',
  'fst', 'snd', 'head', 'tail', 'last', 'skip', 'take', 'take_while', 'drop_while',
  'at', 'pick', 'pad',
  'batch', 'split', 'shard', 'lazy_batch', 'lazy_shard',
  'ndmap', 'ndrange', 'ndenumerate', 'ndflat',
  'transpose', 'transpose_ragged',
  'map', 'filter', 'max', 'min', 'sorted',
  'find_idx', 'find_last_idx', 'find', 'find_last',
  'Iter', 'lift',
]



from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Callable, Generic, TypeVar
from abc import ABC, abstractmethod
from .applicative import Applicative

A = TypeVar('A', covariant=True)
B = TypeVar('B')


class Monad(Applicative[A], ABC, Generic[A]):
  
  @classmethod
  @abstractmethod
  def of(cls, value: B) -> 'Monad[B]':
    ...

  @abstractmethod
  def bind(self, f: Callable[[A], 'Monad[B]']) -> 'Monad[B]':
    ...

  def ap(self, f: 'Monad[Callable[[A], B]]') -> 'Monad[B]':
    return f.bind(lambda func: self.bind(lambda x: self.of(func(x))))
  
  def fmap(self, f: Callable[[A], B]) -> 'Monad[B]':
    def lifted(a: A):
      b = f(a)
      return self.of(b)
    return self.bind(lifted)

  @classmethod
  def pure(cls, value: B) -> 'Monad[B]':
    return cls.of(value)


from haskellian import DEBUG_IMPORTS, asyn_iter as AI
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, AsyncIterable
import asyncio

A = TypeVar('A')

@AI.lift
def prefetched(prefetch: int, xs: AsyncIterable[A]) -> AsyncIterable[A]:
    buffer = asyncio.Queue(maxsize=max(prefetch, 1))
    sentinel = object()

    async def producer():
      async for x in xs:
        await buffer.put(x)
      await buffer.put(sentinel)

    async def consumer():
      while True:
        item = await buffer.get()
        if item is sentinel:
          break
        yield item

    asyncio.create_task(producer())
    return consumer()


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Generic, TypeVar, Literal, Callable, Any, overload
from dataclasses import dataclass
from abc import ABC, abstractmethod
from haskellian import Monad

A = TypeVar('A')
L = TypeVar('L', covariant=True)
R = TypeVar('R', covariant=True)
L2 = TypeVar('L2')
R2 = TypeVar('R2')

@dataclass(eq=False)
class IsLeft(BaseException, Generic[L]):
  value: L

class EitherBase(Monad[R], ABC, Generic[L, R]):

  @abstractmethod
  def _match(self, on_left: Callable[[L], A], on_right: Callable[[R], A]) -> A:
    """Unwrap an `Either` by matching both branches"""
  
  def match(self, on_left: Callable[[L], A], on_right: Callable[[R], A]) -> A:
    """Unwrap an `Either` by matching both branches"""
    return self._match(on_left, on_right)

  @abstractmethod
  def unsafe(self) -> R:
    """Unwraps the value or throws an `IsLeft` exception
    
    (`IsLeft.value` will contain the wrapped value)"""

  def expect(self, error_msg) -> R:
    """Unwraps the value or throws `IsLeft(error_msg)` exception"""
    return self.mapl(lambda _: error_msg).unsafe()
    
  @classmethod
  def of(cls, x: A) -> 'Right[Any, A]':
    return Right(x)

  def bind(self, f: 'Callable[[R], Either[L2, R2]]') -> 'Either[L|L2, R2]':
    return self.match(lambda x: Left(x), f)

  def fmap(self, f: Callable[[R], R2]) -> 'Either[L, R2]':
    return super().fmap(f) # type: ignore

  def mapl(self, f: Callable[[L], L2]) -> 'Either[L2, R]':
    """Map the left branch"""
    return self.match(lambda x: Left(f(x)), lambda x: Right(x))
  
  @overload
  def get_or(self, fallback: R) -> R: ... # type: ignore
  @overload
  def get_or(self, fallback: A) -> A | R: ...
  def get_or(self, fallback): # type: ignore
    return self.match(lambda _: fallback, lambda x: x)

  def ap(self, f: 'Either[L, Callable[[R], R2]]') -> 'Either[L, R2]':
    return super().ap(f) # type: ignore
  
  def __or__(self, f: Callable[[R], R2]) -> 'Either[L, R2]':
    return self.fmap(f)
  
  def match_(self, on_left: Callable[[], A], on_right: Callable[[], A]) -> A:
    """Like `match`, but handlers don't get the wrapped value"""
    return self.match(lambda _: on_left(), lambda _: on_right())
  
@dataclass
class Left(EitherBase[L, R], Generic[L, R]):
  value: L = None # type: ignore
  tag: Literal['left'] = 'left'

  def _match(self, on_left, on_right):
    return on_left(self.value)
  
  def unsafe(self):
    raise IsLeft(self.value)

@dataclass
class Right(EitherBase[L, R], Generic[L, R]):
  value: R = None # type: ignore
  tag: Literal['right'] = 'right'

  def _match(self, on_left, on_right):
    return on_right(self.value)
  
  def unsafe(self) -> R:
    return self.value

Either = Left[L, R] | Right[L, R]


from haskellian import DEBUG_IMPORTS, asyn_iter as AI, Monad, Pipe
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Callable, Generic, TypeVar, Awaitable, AsyncIterator, AsyncIterable, TypeGuard, overload, TypeVarTuple

A = TypeVar('A', covariant=True)
B = TypeVar('B')
As = TypeVarTuple('As')

class AsyncIter(Monad[A], Generic[A], AsyncIterator[A]):

  def __repr__(self):
    return f'AsyncIter({self._xs})'

  def __init__(self, xs: AsyncIterable[A]):
    self._xs = xs

  @classmethod
  def of(cls, value: B) -> 'AsyncIter[B]':
    return AsyncIter(AI.asyncify([value]))

  async def __anext__(self) -> A:
    async for x in self._xs:
      return x
    raise StopAsyncIteration()
  
  def map(self, f: Callable[[A], B]) -> 'AsyncIter[B]':
    return AI.map(f, self)
  
  __or__ = map
  
  def amap(self, f: Callable[[A], Awaitable[B]]) -> 'AsyncIter[B]':
    return AI.amap(f, self)
  
  def bind(self, f: Callable[[A], AsyncIterable[B]]) -> 'AsyncIter[B]':
    return AI.flatmap(f, self)
  
  def flatmap(self, f: Callable[[A], AsyncIterable[B]]) -> 'AsyncIter[B]':
    """Alias of `bind`"""
    return self.bind(f)
  
  def __and__(self, f: Callable[[A], AsyncIterable[B]]) -> 'AsyncIter[B]':
    """Alias of `bind`"""
    return self.bind(f)
  
  @overload
  def filter(self, p: Callable[[A], TypeGuard[B]]) -> 'AsyncIter[B]': ...
  @overload
  def filter(self, p: Callable[[A], bool]) -> 'AsyncIter[A]': ...
  def filter(self, p): # type: ignore
    return AI.filter(p, self)

  async def sync(self) -> list[A]:
    return await AI.syncify(self)
  
  async def head(self) -> A | None:
    return await AI.head(self)
    
  def take(self, n: int) -> 'AsyncIter[A]':
    return AI.take(n, self)
  
  def skip(self, n: int) -> 'AsyncIter[A]':
    return AI.skip(n, self)
  
  def batch(self, n: int) -> 'AsyncIter[tuple[A, ...]]':
    return AI.batch(n, self)
  
  def enumerate(self) -> 'AsyncIter[tuple[int, A]]':
    return AI.enumerate(self)
  
  async def split(self, n: int) -> tuple[list[A], 'AsyncIter[A]']:
    head, xs = await AI.split(n, self)
    return head, AsyncIter(xs)
  
  def prefetch(self, n: int) -> 'AsyncIter[A]':
    return AI.prefetched(n, self) if n > 0 else self


  def i(self, f: Callable[['AsyncIter[A]'], AsyncIterable[B]]) -> 'AsyncIter[B]':
    """Apply an arbitrary iterable function"""
    return AsyncIter(f(self))
  
  def f(self, f: Callable[['AsyncIter[A]'], B]) -> Pipe[B]:
    """Apply an arbitrary function into a `Pipe`"""
    return Pipe(f(self))


from .lifting import lift
from .funcs import all, all2d, bind, delay, of, then, wait, run
from .promise import Promise
from .managed import ManagedPromise

__all__ = [
  'all', 'all2d', 'bind', 'delay', 'of', 'then', 'wait', 'run',
  'Promise', 'lift', 'ManagedPromise'
]


import builtins
from typing_extensions import Callable, TypeVar, Iterable, TypeGuard, overload

A = TypeVar('A')
B = TypeVar('B')

def map(f: Callable[[A], B]) -> Callable[[Iterable[A]], Iterable[B]]:
  return lambda xs: builtins.map(f, xs)

@overload
def filter(p: Callable[[A], TypeGuard[B]]) -> Callable[[Iterable[A]], Iterable[B]]: ...
@overload
def filter(p: Callable[[A], bool]) -> Callable[[Iterable[A]], Iterable[A]]: ...
def filter(p): # type: ignore
  return lambda xs: builtins.filter(p, xs)

def sorted(key: Callable[[A], B] | None = None, reverse: bool = False) -> Callable[[Iterable[A]], Iterable[A]]:
    return lambda xs: builtins.sorted(xs, key=key, reverse=reverse) # type: ignore

def max(key: Callable[[A], B] | None = None) -> Callable[[Iterable[A]], A]:
    return lambda xs: builtins.max(xs, key=key, default=None) # type: ignore

def min(key: Callable[[A], B] | None = None) -> Callable[[Iterable[A]], A]:
    return lambda xs: builtins.min(xs, key=key, default=None) # type: ignore



from .config import DEBUG_IMPORTS
from .classes import Functor, Applicative, Monad
from .thunk import Thunk
from .pipe import Pipe
from .either import Either, Left, Right, IsLeft
from .promise import Promise, ManagedPromise
from .iter import Iter
from .asyn_iter import AsyncIter, ManagedAsync
from .dicts import Dict

__all__ = [
  'DEBUG_IMPORTS',
  'Functor', 'Applicative', 'Monad',
  'Thunk', 'Pipe',
  'Either', 'Left', 'Right', 'IsLeft',
  'Promise', 'ManagedPromise',
  'Iter', 'Dict',
  'AsyncIter', 'ManagedAsync'
]



from typing_extensions import TypeVar, Mapping, overload

D1 = TypeVar('D1', bound=Mapping)
D2 = TypeVar('D2', bound=Mapping)
D3 = TypeVar('D3', bound=Mapping)

def take(Params: type[D1], params: D1 | dict) -> D1:
  return { k: params[k] for k in getattr(Params, '__annotations__', {}) if k in params } # type: ignore

@overload
def split(Params1: type[D1], Params2: type[D2], params: D1 | D2) -> tuple[D1, D2]: ...
@overload
def split(Params1: type[D1], Params2: type[D2], Params3: type[D3], params: D1 | D2 | D3) -> tuple[D1, D2, D3]: ...
def split(Params1: type[D1], Params2: type[D2], x3, x4 = None) -> tuple[Mapping, ...]: # type: ignore
  Params = [Params1, Params2] + (x4 and [x3] or [])
  params = x4 or x3
  return tuple({ k: params[k] for k in getattr(P, '__annotations__', {}) if k in params } for P in Params)


from typing import TypeVar, Iterable, Mapping

A = TypeVar('A')
B = TypeVar('B')

def pluck(xs: Iterable[Mapping[A, B]], key: A) -> Iterable[B]:
  """Extract values by key. Skips dicts without it."""
  for x in xs:
    if key in x:
      yield x[key]


from typing_extensions import Callable, TypeVar, ParamSpec, overload
from functools import reduce

A = ParamSpec('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
E = TypeVar('E')
F = TypeVar('F')
G = TypeVar('G')
H = TypeVar('H')
I = TypeVar('I')

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C]
) -> Callable[A, C]: ...

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C],
  f3: Callable[[C], D]
) -> Callable[A, D]: ...

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C],
  f3: Callable[[C], D],
  f4: Callable[[D], E]
) -> Callable[A, E]: ...

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C],
  f3: Callable[[C], D],
  f4: Callable[[D], E],
  f5: Callable[[E], F]
) -> Callable[A, F]: ...

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C],
  f3: Callable[[C], D],
  f4: Callable[[D], E],
  f5: Callable[[E], F],
  f6: Callable[[F], G]
) -> Callable[A, G]: ...

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C],
  f3: Callable[[C], D],
  f4: Callable[[D], E],
  f5: Callable[[E], F],
  f6: Callable[[F], G],
  f7: Callable[[G], H]
) -> Callable[A, H]: ...

@overload
def flow(
  f1: Callable[A, B],
  f2: Callable[[B], C],
  f3: Callable[[C], D],
  f4: Callable[[D], E],
  f5: Callable[[E], F],
  f6: Callable[[F], G],
  f7: Callable[[G], H],
  f8: Callable[[H], I]
) -> Callable[A, I]: ...

def flow(*fs: Callable): # type: ignore
  return lambda x: reduce(lambda x, f: f(x), fs, x)



from typing import Mapping, TypeVar, Callable, overload
from haskellian import dicts as D
from ramda import curry

K = TypeVar('K')
V = TypeVar('V')
K2 = TypeVar('K2')
V2 = TypeVar('V2')

def unpack(d: Mapping[K, V], *keys: K) -> tuple[V, ...]:
  return tuple(d[key] for key in keys)


@overload
def evolve(mappers: Mapping[K, Callable[[V], V2]], xs: Mapping[K, V]) -> 'D.Dict[K, V|V2]': ...
@overload
def evolve(mappers: Mapping[K, Callable[[V], V2]]) -> Callable[[Mapping[K, V]], 'D.Dict[K, V|V2]']: ...
@curry
def evolve(mappers, xs): # type: ignore
  """Applies each mapper to the corresponding key"""
  return D.Dict({ k: mappers.get(k, lambda x: x)(v) for k, v in xs.items() })


@overload
def insert(k: K, v: V, xs: Mapping[K, V]) -> 'D.Dict[K, V]': ...
@overload
def insert(k: K, v: V) -> Callable[[Mapping[K, V]], 'D.Dict[K, V]']: ...
@curry
def insert(k, v, xs): # type: ignore
  return D.Dict({**xs, k: v})

@overload
def remove(k: K, xs: Mapping[K, V]) -> 'D.Dict[K, V]': ...
@overload
def remove(k: K) -> Callable[[Mapping[K, V]], 'D.Dict[K, V]']: ...
@curry
def remove(key, xs): # type: ignore
  return D.Dict({k: v for k, v in xs.items() if k != key})

@overload
def rename(k1: K, k2: K2, xs: Mapping[K, V]) -> 'D.Dict[K|K2, V]': ...
@overload
def rename(k1: K, k2: K2) -> Callable[[Mapping[K, V]], 'D.Dict[K|K2, V]']: ...
@curry
def rename(k1, k2, xs): # type: ignore
  return D.Dict({**remove(k1, xs), k2: xs[k1]} if k1 in xs else xs)


from haskellian import DEBUG_IMPORTS, asyn_iter as AI, promise as P
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, Generic, AsyncIterator

A = TypeVar('A')
B = TypeVar('B')

class ManagedAsync(AI.AsyncIter[A], Generic[A]):

  def __init__(self):
    self.xs: list[A] = []
    self._next = P.ManagedPromise()
    self.ended: bool = False

  def __repr__(self):
    return f'ManagedAsync({self.xs})'

  def push(self, value: A):
    self.xs.append(value)
    if not self._next.resolved:
      self._next.resolve()

  def end(self):
    self.ended = True
    if not self._next.resolved:
      self._next.resolve()

  def clear(self):
    self.xs = []
    self.ended = False

  def __aiter__(self) -> AsyncIterator[A]:
    return self

  async def __anext__(self) -> A:
    if len(self.xs) > 0:
      return self.xs.pop(0)
    elif self.ended:
      raise StopAsyncIteration()
    else:
      await self._next
      self._next = P.ManagedPromise()
      return await self.__anext__()


from .basics import map, map_k, map_v, filter, filter_k, filter_v, flatmap, flatmap_k, flatmap_v
from .grouping import group_by, zip, unzip, aggregate
from .dict import Dict
from .lifting import lift
from .misc import unpack, evolve, insert, remove, rename

__all__ = [
  'map', 'map_k', 'map_v',
  'filter', 'filter_k', 'filter_v',
  'flatmap', 'flatmap_k', 'flatmap_v',
  'group_by', 'zip', 'unzip', 'aggregate',
  'Dict',
  'lift', 'unpack', 'evolve', 'insert', 'remove', 'rename'
]


from haskellian import DEBUG_IMPORTS, asyn_iter as AI
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import AsyncIterable, Awaitable, Callable, TypeVar, Iterable, TypeVarTuple, Unpack

A = TypeVar("A")
B = TypeVar("B")
As = TypeVarTuple("As")

@AI.lift
async def map(f: Callable[[A], B], xs: AsyncIterable[A]) -> AsyncIterable[B]:
  async for x in xs:
    yield f(x)

@AI.lift
async def amap(f: Callable[[A], Awaitable[B]], xs: AsyncIterable[A]) -> AsyncIterable[B]:
  async for x in xs:
    yield await f(x)

@AI.lift
async def starmap(f: Callable[[Unpack[As]], B], xs: AsyncIterable[tuple[Unpack[As]]]) -> AsyncIterable[B]:
  async for x in xs:
    yield f(*x)

@AI.lift
async def flatmap(f: Callable[[A], AsyncIterable[B]], xs: AsyncIterable[A]) -> AsyncIterable[B]:
  async for x in xs:
    async for y in f(x):
      yield y

@AI.lift
async def filter(p: Callable[[A], bool], xs: AsyncIterable[A]) -> AsyncIterable[A]:
  async for x in xs:
    if p(x):
      yield x

@AI.lift
async def flatten(xxs: AsyncIterable[Iterable[A]]) -> AsyncIterable[A]:
  async for xs in xxs:
    for x in xs:
      yield x

@AI.lift
async def enumerate(xs: AsyncIterable[A]) -> AsyncIterable[tuple[int, A]]:
  i = 0
  async for x in xs:
    yield i, x
    i += 1

@AI.lift
async def asyncify(xs: Iterable[A]) -> AsyncIterable[A]:
  for x in xs:
    yield x

async def syncify(xs: AsyncIterable[A]) -> list[A]:
  return [x async for x in xs]

async def head(xs: AsyncIterable[A]) -> A | None:
  async for x in xs:
    return x

@AI.lift
async def skip(n: int, xs: AsyncIterable[A]) -> AsyncIterable[A]:
  async for i, x in enumerate(xs):
    if i >= n:
      yield x

@AI.lift
async def take(n: int, xs: AsyncIterable[A]) -> AsyncIterable[A]:
  if n == 0:
    return
  async for i, x in enumerate(xs):
    if i < n - 1:
      yield x
    elif i >= n - 1:
      yield x
      return

async def split(n: int, xs: AsyncIterable[A]) -> tuple[list[A], AsyncIterable[A]]:
  head = await syncify(take(n, xs))
  return head, xs

@AI.lift
async def batch(
  batch_size: int, xs: AsyncIterable[A], yield_remaining: bool = True
) -> AsyncIterable[tuple[A, ...]]:
  batch = []
  async for x in xs:
    if len(batch) == batch_size:
      yield tuple(batch)
      batch = []
    batch.append(x)
  if yield_remaining and len(batch) > 0:
    yield tuple(batch)



from typing_extensions import TypeVar
from .either import Either
from .funcs import safe
from pydantic import BaseModel, ValidationError

T = TypeVar('T', bound=BaseModel)

def validate(data, Model: type[T]) -> Either[ValidationError, T]:
  return safe(lambda: Model.model_validate(data), ValidationError)

def validate_json(data: str | bytes | bytearray, Model: type[T]) -> Either[ValidationError, T]:
  return safe(lambda: Model.model_validate_json(data), ValidationError)



from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Callable, Generic, TypeVar
from abc import ABC, abstractmethod

A = TypeVar('A', covariant=True)
B = TypeVar('B')

class Functor(ABC, Generic[A]):
  @abstractmethod
  def fmap(self, f: Callable[[A], B]) -> 'Functor[B]':
    ...

  def __or__(self, f: Callable[[A], B]) -> 'Functor[B]':
    return self.fmap(f)



from haskellian import dicts as D
from typing_extensions import ParamSpec, Callable, Mapping, TypeVar
from functools import wraps

K = TypeVar('K')
V = TypeVar('V')
P = ParamSpec('P')

def lift(func: Callable[P, Mapping[K, V]]) -> Callable[P, D.Dict[K, V]]:
  """Lift a mapping function `func` to return a `Dict`"""
  @wraps(func)
  def _f(*args: P.args, **kwargs: P.kwargs) -> D.Dict[K, V]:
    return D.Dict(func(*args, **kwargs)) # type: ignore pylance wtf
  return _f



from .lifting import lift
from .funcs import amap, asyncify, batch, enumerate, filter, flatmap, flatten, map, skip, split, starmap, syncify, take, head
from .prefetching import prefetched
from .iter import AsyncIter
from .managed import ManagedAsync

__all__ = [
  'amap', 'asyncify', 'batch', 'enumerate', 'filter', 'flatmap', 'flatten', 'map', 'skip', 'split', 'starmap', 'syncify', 'take', 'head',
  'prefetched', 'AsyncIter', 'ManagedAsync', 'lift'
]



from typing_extensions import Iterable, TypeVar
from itertools import zip_longest

A = TypeVar('A')

def transpose(xs: Iterable[Iterable[A]]) -> list[list[A]]:
    """Transpose a 2d list, cropping to the shortest row. E.g:
    ```
    transpose([
        [1, 2],
        [3, 4],
        [5, 6]
    ]) == [
        [1, 3, 5],
        [2, 4, 6]
    ]
    ```
    but
    ```
    transpose([
        [1, 2],
        [3, 4],
        [5]
    ]) == [
        [1, 2, 3],
    ]
    ```
    """
    return list(zip(*xs))

def transpose_ragged(xs: Iterable[Iterable[A]]) -> list[list[A]]:
    """Like `transpose`, but resulting rows can be ragged (i.e. have different lengths). E.g:
    ```
    transpose([
        [1, 2],
        [3, 4],
        [5]
    ]) == [
        [1, 3, 5],
        [2, 4]
    ]
    ```"""
    return [[x for x in cols if x is not None] for cols in zip_longest(*xs)]


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Iterable, TypeVar

A = TypeVar('A')

def at(i: int, xs: list[int]) -> int | None:
  """Safe list indexing"""
  try:
    return xs[i]
  except IndexError:
    ...

def pick(indices: Iterable[int], xs: list[A]) -> list[A]:
  """Pick `xs` values at indices `indices`
  - E.g. `pick([1, 2, 4], ['a', 'b', 'c', 'd', 'e', 'f', 'g']) = ['b', 'c', 'e']`"""
  return [xs[i] for i in indices]


from haskellian import DEBUG_IMPORTS, Monad, promise as P
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, Generic, Callable, Awaitable

A = TypeVar('A', covariant=True)
B = TypeVar('B')

class Promise(Monad[A], Generic[A], Awaitable[A]):

  def __repr__(self) -> str:
    return 'Promise()'
  
  def __init__(self, x: Awaitable[A]):
    self.x = x

  def __await__(self):
    return self.x.__await__()
  
  @classmethod
  def of(cls, x: B) -> 'Promise[B]':
    return Promise(P.of(x))
  
  def then(self, f: Callable[[A], B]) -> 'Promise[B]':
    """Alias for `fmap`"""
    return Promise(P.then(f, self))
  
  def bind(self, f: Callable[[A], Awaitable[B]]) -> 'Promise[B]':
    return Promise(P.bind(f, self))
  
  def fmap(self, f: Callable[[A], B]) -> 'Promise[B]':
    return self.then(f)
  
  def __or__(self, f: Callable[[A], B]) -> 'Promise[B]':
    return self.then(f)

  
  def __and__(self, f: Callable[[A], Awaitable[B]]) -> 'Promise[B]':
    return self.bind(f)
  
  async def run(self) -> A:
    return await self


from haskellian import iter as I
from typing_extensions import ParamSpec, Callable, Iterable, TypeVar
from functools import wraps

A = TypeVar('A')

P = ParamSpec('P')
def lift(func: Callable[P, Iterable[A]]) -> Callable[P, I.Iter[A]]:
  """Lift an iterable function `func` to return an `Iter`"""
  @wraps(func)
  def _f(*args: P.args, **kwargs: P.kwargs) -> I.Iter[A]:
    return I.Iter(func(*args, **kwargs))
  return _f


from haskellian import DEBUG_IMPORTS, iter as I
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import Iterable, TypeVar, overload

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
E = TypeVar('E')
F = TypeVar('F')

@overload
def unzip(xs: Iterable[tuple[A, B]]) -> tuple[list[A], list[B]]: ...
@overload
def unzip(xs: Iterable[tuple[A, B, C]]) -> tuple[list[A], list[B], list[C]]: ...
@overload
def unzip(xs: Iterable[tuple[A, B, C, D]]) -> tuple[list[A], list[B], list[C], list[D]]: ...
@overload
def unzip(xs: Iterable[tuple[A, B, C, D, E]]) -> tuple[list[A], list[B], list[C], list[D], list[E]]: ...
@overload
def unzip(xs: Iterable[tuple[A, B, C, D, E, F]]) -> tuple[list[A], list[B], list[C], list[D], list[E], list[F]]: ...
@overload
def unzip(xs: Iterable[tuple[A, ...]]) -> tuple[list[A], ...]: ...
def unzip(xs):
    """`[(a, b, ...)] -> ([a], [b], ...)`"""
    return tuple(map(list, zip(*xs))) # type: ignore

def uncons(xs: Iterable[A]) -> tuple[A | None, I.Iter[A]]:
    """`uncons([x, *xs]) = (x, xs)`"""
    it = iter(xs)
    x = next(it, None)
    return x, I.Iter(it)

@I.lift
def pairwise(xs: Iterable[A]) -> Iterable[tuple[A, A]]:
    """`pairwise([x1, x2, x3, x4, ...]) = [(x1, x2), (x2, x3), (x3, x4), ...]`
    - `pairwise([]) = []`
    - `pairwise([x]) = []`
    """
    x0, tail = uncons(xs)
    for x1 in tail:
        yield x0, x1 # type: ignore (`x0` won't be `None` if there's some element in `tail`)
        x0 = x1



from haskellian import DEBUG_IMPORTS, asyn_iter as AI
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import ParamSpec, Callable, AsyncIterable, TypeVar
from functools import wraps

A = TypeVar('A')

P = ParamSpec('P')
def lift(f: Callable[P, AsyncIterable[A]]) -> Callable[P, AI.AsyncIter[A]]:
  """Lift a function `f` to return an `AsyncIter`"""
  @wraps(f)
  def _f(*args: P.args, **kwargs: P.kwargs) -> AI.AsyncIter[A]:
    return AI.AsyncIter(f(*args, **kwargs))
  return _f


from typing_extensions import Callable, TypeVar, ParamSpec, Generic, overload, Coroutine, Awaitable
from inspect import iscoroutinefunction
from functools import wraps

T = TypeVar('T')
P = ParamSpec('P')

class IsNone(BaseException):
  ...

def unsafe(x: T | None) -> T:
  if x is None:
    raise IsNone()
  return x

@overload
def do(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
  ...
@overload
def do(fn: Callable[P, T]) -> Callable[P, T | None]:
  ...
def do(fn): # type: ignore
  """Lift a function with do notation
  ```
  @do
  def safe_sum(a: int | None, b: int | None):
    return O.unsafe(a) + O.unsafe(b)

  safe_sum(1, 2) # 3
  safe_sum(None, 2) # None
  safe_sum(1, None) # None
  safe_sum(None, None) # None
  ```
  """
  if iscoroutinefunction(fn):
    @wraps(fn)
    async def _wrapper(*args: P.args, **kwargs: P.kwargs):
      try:
        return await fn(*args, **kwargs)
      except IsNone:
        return None
    return _wrapper
  else:
    @wraps(fn)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
      try:
        return fn(*args, **kwargs)
      except IsNone:
        return None
    return wrapper


from typing import TypeVar, Mapping, Callable, TypeGuard, overload
import inspect
from haskellian import dicts as D

K = TypeVar('K')
V = TypeVar('V')
K2 = TypeVar('K2')
V2 = TypeVar('V2')
T = TypeVar('T')


@D.lift
def map(f: Callable[[K, V], V2], xs: Mapping[K, V]) -> dict[K, V2]:
  return { k: f(k, v) for k, v in xs.items() }

@D.lift
def map_v(f: Callable[[V], V2], xs: Mapping[K, V]) -> dict[K, V2]:
  return { k: f(v) for k, v in xs.items() }

@D.lift
def map_k(f: Callable[[K], K2], xs: Mapping[K, V]) -> dict[K2, V]:
  return { f(k): v for k, v in xs.items() }

@D.lift
def flatmap(f: Callable[[K, V], Mapping[K2, V2]], xs: Mapping[K, V]) -> dict[K2, V2]:
  out: dict[K2, V2] = {}
  for entry in xs.items():
    out |= f(*entry)
  return out

@D.lift
def flatmap_v(f: Callable[[V], Mapping[K2, V2]], xs: Mapping[K, V]) -> dict[K2, V2]:
  out: dict[K2, V2] = {}
  for _, v in xs.items():
    out |= f(v)
  return out

@D.lift
def flatmap_k(f: Callable[[K], Mapping[K2, V]], xs: Mapping[K, V]) -> dict[K2, V]:
  out: dict[K2, V] = {}
  for k in xs:
    out |= f(k)
  return out

@D.lift
def filter(f: Callable[[K, V], bool], xs: Mapping[K, V]) -> dict[K, V]:
  return { k: v for k, v in xs.items() if f(k, v) }

@D.lift
def filter_v(f: Callable[[V], bool], xs: Mapping[K, V]) -> dict[K, V]:
  return { k: v for k, v in xs.items() if f(v) }

@D.lift
def filter_k(f: Callable[[K], bool], xs: Mapping[K, V]) -> dict[K, V]:
  return { k: v for k, v in xs.items() if f(k) }


from typing import Mapping, TypeVar, Callable, Iterable, Sequence, TypeGuard, overload
from haskellian import iter as I, dicts as D

A = TypeVar('A')
K = TypeVar('K')
K1 = TypeVar('K1')
K2 = TypeVar('K2')
V = TypeVar('V')
V1 = TypeVar('V1')
V2 = TypeVar('V2')

@D.lift
def group_by(f: Callable[[V], K], xs: Iterable[V]) -> dict[K, list[V]]:
  """Group elements by a function"""
  from collections import defaultdict
  d = defaultdict(list)
  for x in xs:
    d[f(x)].append(x)
  return dict(d)

@I.lift
def zip(xs: Mapping[K, Iterable[V]]) -> Iterable['D.Dict[K, V]']:
  """`zip({ 'a': [1, 2, 3], 'b': [4, 5, 6] }) == [{'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 3, 'b': 6}]`"""
  iters = {
    key: iter(it)
    for key, it in xs.items()
  }
  while True:
    try:
      yield D.Dict({
        key: next(it)
        for key, it in iters.items()
      })
    except StopIteration:
      break

@D.lift
def unzip(xs: Sequence[Mapping[K, V]]) -> dict[K, list[V]]:
  """`unzip([{'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 3, 'b': 6}]) == { 'a': [1, 2, 3], 'b': [4, 5, 6] }`"""
  from collections import defaultdict
  out = defaultdict(list)
  for x in xs:
    for key, value in x.items():
      out[key].append(value)
  return dict(out)

@D.lift
def aggregate(f: Callable[[Sequence[V]], A], xs: Sequence[Mapping[K, V]]) -> dict[K, A]:
  """Aggregate values with a same key
  ```
  aggregate(sum, [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])
  # {'a': 4, 'b': 6}
  ```
  """
  return {
    key: f(list(I.pluck(xs, key)))
    for key in xs[0].keys()
  }




from typing_extensions import Iterable, TypeVar, Callable
import itertools
from haskellian import iter as I, Iter

A = TypeVar('A')

def split(n: int, xs: Iterable[A]) -> tuple[list[A], Iterable[A]]:
  """Splits `xs` into a list of the first `n` and a generator of the rest
  - e.g: `split(3, [1,2,3,4,5]) == ([1, 2, 3], generator(4, 5))`"""
  it = iter(xs)
  init = I.take(n, it).sync()
  return init, (x for x in it)

@I.lift
def batch(n: int, xs: Iterable[A]) -> Iterable[tuple[A, ...]]:
  """Batches `xs` into `n`-tuples"""
  it = iter(xs)
  while b := tuple(itertools.islice(it, n)):
    yield b

@I.lift
def lazy_batch(n: int, xs: Iterable[A]) -> Iterable[Iter[A]]:
  """Batches `xs` into `n`-tuples, but returns lazy iterators"""
  it = iter(xs)
  while not (b := I.take(n, it)).empty():
    yield b

@I.lift
def shard(min_size: float, size: Callable[[A], float], xs: Iterable[A]) -> Iterable[list[A]]:
  """Shards `xs` into groups of at least `min_size` based on `size` (last shard may have less)"""
  shard = []
  acc = 0
  for x in xs:
    acc += size(x)
    shard.append(x)
    if acc >= min_size:
      yield shard
      shard = []
      acc = 0
  if shard:
    yield shard

@I.lift
def lazy_shard(min_size: float, size: Callable[[A], float], xs: Iterable[A]) -> Iterable[Iter[A]]:
  """Shards `xs` into groups of at least `min_size` based on `size` (last shard may have less)"""
  it = iter(xs)
  @I.lift
  def shard():
    acc = 0
    for x in it:
      acc += size(x)
      yield x
      if acc >= min_size:
        break
  while not (s := shard()).empty():
    yield s


from haskellian import DEBUG_IMPORTS, promise as P
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import ParamSpec, Callable, Awaitable, TypeVar
from functools import wraps

A = TypeVar('A')
Ps = ParamSpec('Ps')

def lift(f: Callable[Ps, Awaitable[A]]) -> Callable[Ps, P.Promise[A]]:
  """Lift an async function `f` to return a `Promise`"""
  @wraps(f)
  def _f(*args: Ps.args, **kwargs: Ps.kwargs) -> P.Promise[A]:
    return P.Promise(f(*args, **kwargs))
  return _f


from haskellian import DEBUG_IMPORTS, iter as I
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, Literal, Callable, Iterable, overload, Any
from .basics import isiterable

A = TypeVar('A')
B = TypeVar('B')

@overload
def ndmap(f: Callable[[A], B], xs: Iterable[Iterable[Iterable[A]]], depth: Literal[3]) -> list[list[list[B]]]: ...
@overload
def ndmap(f: Callable[[A], B], xs: Iterable[Iterable[Iterable[Iterable[A]]]], depth: Literal[4]) -> list[list[list[list[B]]]]: ...
@overload
def ndmap(f: Callable[[A], B], xs: Iterable[Iterable[A]], depth: Literal[2] = 2) -> list[list[B]]: ...
@overload
def ndmap(f: Callable[[A], B], xs: Iterable, depth: int) -> list: ...
def ndmap(f, xs, depth = 2):
  return f(xs) if depth == 0 else [ndmap(f, x, depth-1) for x in xs]

Range = int | tuple[int, int] | tuple[int, int, int]
@overload
def ndrange(rng1: Range) -> I.Iter[tuple[int]]: ...
@overload
def ndrange(rng1: Range, rng2: Range) -> I.Iter[tuple[int, int]]: ...
@overload
def ndrange(rng1: Range, rng2: Range, rng3: Range) -> I.Iter[tuple[int, int, int]]: ...
@overload
def ndrange(*ranges: Range) -> I.Iter[tuple[int, ...]]: ...
@I.lift
def ndrange(*ranges): # type: ignore
  """Like `range`, but returns an iterable of `n`-tuples instead of ints.
  
  So, instead of
  ```
  for i in range(m):
  for j in range(n):
    ...
  ```
  You can just do:
  ```
  for i, j in ndrange(m, n):
  ...
  ```
  
  And it also supports normal `(start, end)` or `(start, end, step)` ranges:
  ```
  for i, j in ndrange((-10, 0, 2), (4, 6)):
  ...
  # (-10, 4), (-10, 5), (-8, 4), (-8, 5), ...
  ```
  """
  def _ndrange(*ranges, inds = ()):
    if len(ranges) == 0:
      yield inds
    else:
      [r, *rs] = ranges
      rng = range(*r) if isiterable(r) else range(r)
      for i in rng:
        yield from _ndrange(*rs, inds=inds+(i,))
  
  yield from _ndrange(*ranges)

@I.lift
def _fixed_ndenumerate(xxs, depth: int, inds: tuple[int, ...] = ()):
  if depth == 0:
    yield inds, xxs
  else:
    for i, xs in enumerate(xxs):
      yield from _fixed_ndenumerate(xs, depth-1, inds + (i, ))

@I.lift
def _auto_ndenumerate(xxs, inds: tuple[int, ...] = ()):
  if not isiterable(xxs):
    yield inds, xxs
  else:
    for i, xs in enumerate(xxs):
      yield from _auto_ndenumerate(xs, inds + (i,))

@overload
def ndenumerate(xxs: Iterable[Iterable[A]], depth: Literal[2] | None = None) -> I.Iter[tuple[tuple[int, int], A]]: ...
@overload
def ndenumerate(xxs: Iterable[Iterable[Iterable[A]]], depth: Literal[3] | None = None) -> I.Iter[tuple[tuple[int, int, int], A]]: ...
@overload
def ndenumerate(xxs: Iterable, depth: int | None = None) -> I.Iter[tuple[tuple[int, ...], Any]]: ...
@I.lift
def ndenumerate(xxs, depth: int | None = None): # type: ignore
  """Like `enumerate`, but for nested iterables.
  
  So, instead of
  
  ```
  for i, xxs in enumerate(xxxs):
  for j, xs in enumerate(xxs):
    for k, x in enumerate(xs):
    ...
  ```
  
  You can just do
  
  ```
  for x, (i, j, k) in ndenumerate(xxxs):
  ...
  ```
  
  Note:
  - By default it unrolls all the way down until a none-iterable object is found.
  - If you want to iterate over iterables (e.g. lists, numpy arrays, tensors), you'll have to pass in a fixed `depth`
  """
  return _auto_ndenumerate(xxs) if depth is None else _fixed_ndenumerate(xxs, depth)

def _fixed_ndflat(xxs, depth: int):
  if depth == 0:
    yield xxs
  else:
    for xs in xxs:
      yield from _fixed_ndflat(xs, depth-1)
      
def _auto_ndflat(xxs):
  if not isiterable(xxs):
    yield xxs
  else:
    for xs in xxs:
      yield from _auto_ndflat(xs)
      
@overload
def ndflat(xxs: Iterable[Iterable[A]], depth: Literal[2] | None = None) -> I.Iter[I.Iter[A]]: ...
@overload
def ndflat(xxs: Iterable[Iterable[Iterable[A]]], depth: Literal[3] | None = None) -> I.Iter[A]: ...
@overload
def ndflat(xxs: Iterable, depth: int | None = None) -> I.Iter: ...
@I.lift
def ndflat(xxs: Iterable[Iterable[A]], depth: int | None = None): # type: ignore
  """`ndflat([[x1, x2], [x3, [x4]]]) = [x1, x2, x3, x4]`"""
  return _auto_ndflat(xxs) if depth is None else _fixed_ndflat(xxs, depth) # type: ignore


from haskellian import DEBUG_IMPORTS, promise as P
if DEBUG_IMPORTS:
  print('Import:', __name__)
from typing_extensions import TypeVar, Generic
import asyncio

A = TypeVar('A', covariant=True)
B = TypeVar('B')

class ManagedPromise(P.Promise[A], Generic[A]):
  
  value: A

  def __init__(self):
    self.resolved = False
    self.event = asyncio.Event()

  def resolve(self, value: A = None): # type: ignore
    self.value = value
    self.resolved = True
    self.event.set()

  def __await__(self):
    async def wait():
      await self.event.wait()
      return self.value
    return wait().__await__()
  


from dataclasses import dataclass
from typing_extensions import Generic, TypeVar, Callable, TypeGuard, Iterator, Iterable, overload, Any, Literal
from haskellian import iter as I, Monad, Pipe

A = TypeVar('A', covariant=True)
B = TypeVar('B')

@dataclass
class Iter(Monad[A], Iterator[A], Generic[A]):
  
  def __init__(self, xs: Iterable[A] = []):
    self.xs = (x for x in xs)

  def __repr__(self):
    previewed, rest = I.split(5, self.xs)
    self.xs = I.flatten([previewed, rest])
    if len(previewed) == 0:
      return 'Iter([])'
    elif len(previewed) < 5:
      return f'Iter({previewed})'
    else:
      return f'Iter([{", ".join(str(x) for x in previewed)}, ...])'
    
  def empty(self) -> bool:
    x, xs = I.split(1, self.xs)
    self.xs = I.flatten([x, xs])
    return len(x) == 0

  def __next__(self) -> A:
    for x in self.xs:
      return x
    raise StopIteration()

  @classmethod
  def of(cls, value: B) -> 'Iter[B]':
    return Iter([value])
  
  def bind(self, f: Callable[[A], Iterable[B]]) -> 'Iter[B]':
    return I.flatmap(f, self)
  
  def flatmap(self, f: Callable[[A], Iterable[B]] = lambda x: x) -> 'Iter[B]': # type: ignore
    """Acts like `flatten` by default (`f` defaults to the identity)"""
    return self.bind(f)
  
  def iflatmap(self, f: Callable[[int, A], Iterable[B]]) -> 'Iter[B]':
    """`flatmap` with indices"""
    return self.enumerate().flatmap(lambda xs: f(*xs))
  
  def map(self, f: Callable[[A], B]) -> 'Iter[B]':
    return Iter(map(f, self))
  
  fmap = map

  def __or__(self, f: Callable[[A], B]) -> 'Iter[B]':
    """Alias of `map`"""
    return self.map(f)
  
  @overload
  def filter(self, p: Callable[[A], TypeGuard[B]]) -> 'Iter[B]': ...
  @overload
  def filter(self, p: Callable[[A], bool]) -> 'Iter[A]': ...
  def filter(self, p): # type: ignore
    return Iter(filter(p, self))

  @overload
  def reduce(self, f: Callable[[B, A], B], init: B) -> B: ...
  @overload
  def reduce(self, f: Callable[[A, A], A]) -> A | None: ...

  def reduce(self, f, init=None): # type: ignore
    from functools import reduce
    if init is None:
      init = self.head()
      if init is not None:
        return reduce(f, self, init)
    else:
      return reduce(f, self, init)

  @overload
  def batch(self, n: int) -> 'Iter[tuple[A, ...]]': ...
  @overload
  def batch(self, n: int, *, lazy: Literal[True]) -> 'Iter[Iter[A]]': ...
  def batch(self, n: int, *, lazy: bool = False):
    return I.lazy_batch(n, self) if lazy else I.batch(n, self)
  
  @overload
  def shard(self, min_size: float, size: Callable[[A], float]) -> 'Iter[list[A]]': ...
  @overload
  def shard(self, min_size: float, size: Callable[[A], float], *, lazy: Literal[True]) -> 'Iter[Iter[A]]': ...
  def shard(self, min_size: float, size: Callable[[A], float], *, lazy: bool = False):
    """Shards `self` into groups of at least `min_size` based on `size` (last shard may have less)"""
    return I.lazy_shard(min_size, size, self) if lazy else I.shard(min_size, size, self)
  
  def head(self) -> A | None:
    return I.head(self)
  
  def tail(self) -> 'Iter[A]':
    return I.tail(self)
  
  def uncons(self) -> tuple[A | None, 'Iter[A]']:
    return I.uncons(self)
  
  def split(self, n: int) -> 'tuple[Iter[A], Iter[A]]':
    xs, ys = I.split(n, self)
    return Iter(xs), Iter(ys)
  
  def take(self, n: int) -> 'Iter[A]':
    return I.take(n, self)
  
  def skip(self, n: int) -> 'Iter[A]':
    return I.skip(n, self)
  
  def take_while(self, f: Callable[[A], bool]) -> 'Iter[A]':
    return I.take_while(f, self)
  
  def at(self, i: int) -> A | None:
    return self.skip(i).head()

  def pairwise(self) -> 'Iter[tuple[A, A]]':
    return I.pairwise(self)
  
  def enumerate(self) -> 'Iter[tuple[int, A]]':
    return Iter(enumerate(self))
  
  def sort(self, key: Callable[[A], Any] | None = None, reverse: bool = False) -> 'Iter[A]':
    """Like `sorted`, but returns an `Iter`"""
    return Iter(sorted(self, key=key, reverse=reverse)) # type: ignore
  
  def sorted(self, key: Callable[[A], Any] | None = None, reverse: bool = False) -> list[A]:
    return sorted(self, key=key, reverse=reverse) # type: ignore
  
  def min(self, key: Callable[[A], Any] | None = None) -> A | None:
    return min(self, key=key, default=None) # type: ignore
  
  def max(self, key: Callable[[A], Any] | None = None) -> A | None:
    return max(self, key=key, default=None) # type: ignore
  
  def find_idx(self, p: Callable[[A], bool]) -> int | None:
    return I.find_idx(p, self)
  
  def find(self, p: Callable[[A], bool]) -> A | None:
    return I.find(p, self)
  
  def find_last_idx(self, p: Callable[[A], bool]) -> int | None:
    return I.find_last_idx(p, list(self))
  
  def find_last(self, p: Callable[[A], bool]) -> A | None:
    return I.find_last(p, list(self))

  def tap(self, f: Callable[[A], Any]) -> 'Iter[A]':
    return I.tap(f, self)

  def pad(self, n: int, fill: B) -> 'Iter[A|B]':
    return I.pad(n, fill, self)

  def sync(self) -> list[A]:
    return list(self.xs)
  
  def len(self) -> int:
    return len(list(self))

  @overload
  def append(self, x: A) -> 'Iter[A]': ... # type: ignore
  @overload
  def append(self, x: B) -> 'Iter[A|B]': ...
  def append(self, x): # type: ignore
    return self.extend([x])

  @overload
  def extend(self, xs: Iterable[A]) -> 'Iter[A]': ...
  @overload
  def extend(self, xs: Iterable[A|B]) -> 'Iter[A|B]': ...
  def extend(self, xs): # type: ignore
    return I.flatten([self, xs])
  
  def i(self, f: Callable[['Iter[A]'], Iterable[B]]) -> 'Iter[B]':
    """Apply an arbitrary iterable function"""
    return Iter(f(self))
  
  def f(self, f: Callable[['Iter[A]'], B]) -> Pipe[B]:
    """Apply an arbitrary function into a `Pipe`"""
    return Pipe(f(self))


from haskellian import DEBUG_IMPORTS
if DEBUG_IMPORTS:
  print('Import:', __name__)
from dataclasses import dataclass
from typing_extensions import TypeVar, Callable, Generic
from haskellian import Monad

A = TypeVar('A')
B = TypeVar('B')


@dataclass
class Pipe(Monad[A], Generic[A]):
  """The simplest monad: just a value. But one that can be easily operated on:
  
  ```
  (Pipe('world hello')
  .f(str.title)
  .f(lambda s: s.split(' '))
  .f(sorted)
  .f(', '.join)
  ).value
  # 'Hello, World'
  ```
  """

  value: A

  def bind(self, f: Callable[[A], 'Pipe[B]']) -> 'Pipe[B]':
    return f(self.value)
  
  def f(self, f: Callable[[A], B]) -> 'Pipe[B]':
    return self.fmap(f) # type: ignore
  
  @classmethod
  def of(cls, value: A) -> 'Pipe[A]':
    return cls(value)